%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Curry2Prolog evaluator of main expressions.
%
% required: environment variable PAKCSHOME

:- module(evaluator,
	  [currentprogram/1, numberOfCalls/1, numberOfExits/1,
	   singlestep/0, tracemode/0, spymode/0, spypoints/1,
	   printDepth/1, printAllFailures/0,
	   profiling/1, suspendmode/1, timemode/1, 
	   evaluateGoalAndExit/1, evaluateMainExpression/3,
	   writeCurry/1, writeVar/2, bindFreeVars/1]).
	  
:- use_module(prologbasics).
:- use_module(basics).
:- use_module(library(system)).
:- prolog(sicstus) -> use_module('lib_src/prim_readshowterm')  % for showing characters
                    ; use_module('libswi/prim_readshowterm').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic numberOfCalls/1, numberOfExits/1, singlestep/0, tracemode/0,
	   spypoints/1, spymode/0, spyFail/0, printDepth/1,
	   profiling/1, profile_data/3, currentprogram/1,
	   suspendmode/1, timemode/1, nextIOproof/0,
	   printAllFailures/0, errorAbort/0.

currentprogram("prelude").
numberOfCalls(0). % number of function calls
numberOfExits(0). % number of function exits (including suspensions)
%nextIOproof. % will be asserted in an alternative proof to check IO non-det.
%errorAbort. % will be asserted in case of an abort causes by error/user
singlestep. % single step mode initially on in debug mode
tracemode.  % trace mode initially on in debug mode
spypoints([]). % list of spy points
%spymode. % initially no spy points
%spyFail. % show fail ports in spy mode
printDepth(0). % maximal print depth of terms
profiling(no). % show profiling statistics in debug mode
suspendmode(no). % yes if suspended goals should be shown
timemode(no).	 % yes if execution times should be shown


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% evaluate a goal and exit (used in saved states):
evaluateGoalAndExit(Goal) :-
	(call(Goal) -> true ; true),
	halt.

% evaluate an expression with a given type and a given list of free variables:
evaluateMainExpression(Exp,Type,Vs) :-
	prompt(_,''), % redefine prompt for trace mode
	retract(numberOfCalls(_)), retract(numberOfExits(_)),
	asserta(numberOfCalls(0)), asserta(numberOfExits(0)),
	retractAllFacts(profile_data/3),
	(retract(nextIOproof) -> true ; true), % clean fact for nextIOproof/0
	clearDynamicPreds,
	worldToken(World),
	(isIoType(Type) -> E=apply(Exp,World) ; E=Exp),
	(retract(hasPrintedFailure) -> true ; true),
	getRunTime(RTime1),
	getElapsedTime(ETime1),
	evalToken(Eval),
	extractMakeShareInTerm(E,MSE),
	on_exception(ErrorMsg,
	             callAndReturnSuspensions(user:nf(MSE,V,Eval,Done),Suspended),
		     (asserta(errorAbort),
		      ErrorMsg=debugger_abort
		       -> write('Execution aborted.'), nl, fail
		        ; print_error(ErrorMsg)) ),
	getRunTime(RTime2),
	getElapsedTime(ETime2),
	% don't print failures after backtracking:
	((hasPrintedFailure ; printAllFailures) -> true
	 ; asserta(hasPrintedFailure)),
	(timemode(yes)
           -> write('Runtime: '), RTime is RTime2-RTime1, write(RTime),
	      write(' msec.'), nl,
	      write('Elapsed time: '), ETime is ETime2-ETime1, write(ETime),
	      write(' msec.'), nl
           ; true),
	numberOfCalls(NC), numberOfExits(NE),
	(NC>0 -> write('Number of function calls: '), write(NC), nl,
	         write('Number of function exits: '), write(NE), nl
               ; true),
	(var(Done)
         -> suspendmode(SM),
	    (SM=yes -> writeErr('*** Goal suspended!')
	     ; writeErr('*** Goal suspended (use ":set +suspend" for details)!')),
            (Vs=[] -> true ; nlErr, write('Bindings: '), %bindFreeVars(Vs),
                             writeBindings(Vs), nl),
            writeSuspendedGoals(Suspended), nl, flush_output,
	    showProfileData,
	    !, fail
         ; ((nonvar(V), V='$io'('()'), nextIOproof)
            -> flush_output, retract(nextIOproof),
               writeErr('ERROR: non-determinism in I/O actions occurred!'),
               nlErr,
	       showProfileData,
	       !, fail
            ; suspendmode(SM),
	      ((SM=yes;Suspended=[]) -> true
	       ; writeErr('*** Warning: there are suspended constraints (for details: ":set +suspend")'), nlErr),
	      ((nonvar(V), V='$io'('()'))
               -> (Vs=[] -> true ; write('Bindings: '), bindFreeVars(Vs),
                                   writeBindings(Vs), nl),
                  flush_output,
		  (profiling(yes) % no IO ND checking during profiling
		   -> showProfileData, !, fail
		    ; (hasPrintedFailure -> true
		       ; asserta(hasPrintedFailure)), % don't print failures
		                                      % during IO ND checking
		      asserta(nextIOproof), fail)
	       ; write('Result: '), bindFreeVars(Vs),
                 writeCurry(V),
	         (Vs=[] ->true ; nl, write('Bindings: '), writeBindings(Vs)),
	         writeSuspendedGoals(Suspended), write(' ? '),
                 flush_output,
	         readMore(More),
		 \+ More = ";",  % backtrack if user types ";"
		 showProfileData,
		 !,
		 % store command (beginning with ":") for processing:
		 (More=[58|_] -> storeFirstCmds([More]) ; true),
		 More = -1))). % do not fail if user types end-of-file
evaluateMainExpression(_,_,_) :-
	% no further message ("no more sols") in case of abort:
	retract(errorAbort),
	!, fail.
evaluateMainExpression(_,_,_) :- % ignore proof attempt for IO ND
	retract(nextIOproof),
	showProfileData,
	!, fail.
evaluateMainExpression(_,_,_) :-
	write('No more solutions.'), nl,
	showProfileData,
	!, fail.

readMore(More) :-
	readLine(Line),
	(Line = -1 -> More=Line
	            ; removeBlanks(Line,Input), processReadMore(Input,More)).

processReadMore(M,More) :-
	pakcsrc(moresolutionsonreturn,yes), !,
	((M="" ; M="n" ; M=[58|_]) % 58=':'
	 -> (M="" -> More=";" ; (M="n" -> More="" ; More=M))
          ; write('Type <return> for more or "n" > '),
	    readMore(More)).
processReadMore(M,More) :-
	((M="" ; M=";" ; M=[58|_]) % 58=':'
	 -> M=More
          ; write('Type ";" for more or <return> > '),
	    readMore(More)).


writeSuspendedGoals(_) :-
	suspendmode(no), !.
writeSuspendedGoals([]) :- !.
writeSuspendedGoals(Suspended) :-
	nl, write('Suspended goals (in internal representation):'),
	writeSuspendedGoalList(Suspended).

writeSuspendedGoalList([]).
writeSuspendedGoalList([_-G|Suspended]) :-
	nl,
	tryWriteSuspGoal(G),
	writeSuspendedGoalList(Suspended).

% try to format suspended goals more nicely:
tryWriteSuspGoal(_:G) :-
	G =.. [Pred|Args],
	rev(Args,[_,_,Result|RArgs]),
	rev(RArgs,FunArgs),
	FunCall =.. [Pred|FunArgs],
	write('let '), writeCurry(Result), write(' = '),
	writeCurry(FunCall),
	!.
tryWriteSuspGoal(_:G) :- !, 
	writeCurry(G).
tryWriteSuspGoal(G) :- write(G).


% extract all makeShare terms and replace them by the result of execute makeShare:
extractMakeShareInTerm(V,V) :- var(V), !.
extractMakeShareInTerm(makeShare(T,V),V) :- !,
	extractMakeShareInTerm(T,MST),
	makeShare(MST,V).
extractMakeShareInTerm(T,NT) :-
	T =.. [F|Args],
	map2(extractMakeShareInTerm,Args,NArgs),
	NT =.. [F|NArgs].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% show profile data, if desired:
showProfileData :-
	%(plprofiling(yes) -> view([user:_]) ; true),
	profiling(yes),
	currentprogram(Prog),
	append(Prog,".profile",ProgP),
	atom_codes(ProgPName,ProgP),
	tryWriteFile(ProgPName),
	!,
	tell(ProgPName),
	write('Profile data:'), nl,
	findall(pdata(F,N,Info),profile_data(F,N,Info),PD),
	sort(PD,SortedPD),
	map1(format_profile_entry,SortedPD),
	nl,
	told,
	append("more ",ProgP,MP),
	atom_codes(Cmd,MP),
        system(Cmd).
showProfileData.

% printing formatted profile data entries:
format_profile_entry(pdata(Func,_,port(C,F,E,R))) :-
	write(Func), write(': '),
	write('Calls='), write(C), write(' / '),
	write('Fails='), write(F), write(' / '),
	write('Exits='), write(E), write(' / '),
	write('Redos='), write(R), nl.

% predicates for storing profile information during debugging:
profileCall(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	C1 is C+1,
	asserta(profile_data(PF,PN,port(C1,F,E,R))).

profileFail(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	F1 is F+1,
	asserta(profile_data(PF,PN,port(C,F1,E,R))).

profileExit(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	E1 is E+1,
	asserta(profile_data(PF,PN,port(C,F,E1,R))).

profileRedo(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	R1 is R+1,
	asserta(profile_data(PF,PN,port(C,F,E,R1))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write a (Prolog) term in Curry syntax:
% (used in curry2prolog for printing results or during debugging)

writeCurry(T) :- printDepth(D), removeShares(T,UT), writeCurryD(D,UT).

writeCurryD(D,_) :- D=1, !, write('...').
writeCurryD(_,T) :- var(T), !, write(T).
writeCurryD(_,T) :- number(T), !, write(T).
writeCurryD(_,T) :- isCharCons(T), !,
	show_term(T,_,TS,[]), % use ReadShowTerm.showTerm for showing chars
	string2Atom(TS,TA), write(TA).
writeCurryD(_,[]) :- write([]), !.
writeCurryD(_,T) :- atom(T), !,
	(atom_codes(T,[95|_]) -> writeVar(user_output,T) % 95 = '_'
	 ; revTransFunctor(T,ExtName),
	   (isId(ExtName) -> write(ExtName)
                           ; write('('), write(ExtName), write(')'))).
writeCurryD(D,T) :-
	isCompleteList(T),
	!,
	(isString(T)
	    -> % use ReadShowTerm.showTerm for showing strings:
	       show_term(T,_,TS,[]), string2Atom(TS,TA), write(TA)
	     ; write('['), writeCurryList(D,T)).
writeCurryD(D,[T|Ts]) :- !,
	write('('), writeCurryConsList(D,[T|Ts]), write(')').
writeCurryD(D,partcall(_,F,Args)) :- !,
	rev(Args,RArgs), Term =.. [F|RArgs], writeCurryD(D,Term).
writeCurryD(D,apply(F,X)) :- !,
	D1 is D-1,
	write('('), writeCurryD(D1,F), write(' '),
	writeCurryD(D1,X), write(')').
writeCurryD(D,if_then_else(C,T,E)) :- !,
	D1 is D-1,
	write('(if '), writeCurryD(D1,C),
	write(' then '), writeCurryD(D1,T),
	write(' else '), writeCurryD(D1,E), write(')').
writeCurryD(D,'prelude..'(F,G)) :- !,
	D1 is D-1,
	write('('), writeCurryD(D1,F),
	write(' . '), writeCurryD(D1,G), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [Cons,Arg1|Args],
	atom_codes(Cons,[40,44|_]), % tuple constructor
	!,
	write('('), writeCurryD(D1,Arg1), writeCurryTuple(D1,Args), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [IntCons,Arg1,Arg2],
	revTransFunctor(IntCons,Cons),
	\+ isId(Cons),		% write as an infix operator:
	!,
	write('('), writeCurryD(D1,Arg1), write(' '),
	write(Cons), write(' '),
	writeCurryD(D1,Arg2), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [IntCons|Args],
	revTransFunctor(IntCons,Cons),
	write('('), write(Cons),
	writeCurryArgs(D1,Args).

writeCurryTuple(_,[]).
writeCurryTuple(D,[A|As]) :-
	write(','),
	writeCurryD(D,A),
	writeCurryTuple(D,As).

writeCurryList(_,[]).
writeCurryList(D,[_|_]) :- D=1, !, write('...]').
writeCurryList(D,[A|As]) :-
	D1 is D-1,
	writeCurryD(D,A),
	(As=[] -> write(']') ; write(',')),
	writeCurryList(D1,As).

writeCurryConsList(D,_) :- D=1, !, write('...').
writeCurryConsList(D,T) :-
	nonvar(T), T = [A|As], !,
	D1 is D-1,
	writeCurryD(D,A),
	write(':'),
	writeCurryConsList(D1,As).
writeCurryConsList(D,T) :- writeCurryD(D,T).

writeCurryArgs(_,[]) :- write(')').
writeCurryArgs(D,[A|As]) :-
	write(' '),
	writeCurryD(D,A),
	writeCurryArgs(D,As).


% translate a functor (operation or constructor symbol) from
% internal into external representation:
revTransFunctor(IntName,Name) :-
	constructorOrFunctionType(IntName,Name,_,_), !.
revTransFunctor(Name,Name).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% write a variable (i.e., omit first char which is always '_'):
writeVar(Str,V) :-
	atom_codes(V,[_|Cs]), atom_codes(V1,Cs),
	write(Str,V1).

% show an output substitution in a readable way:
writeBindings([]).
writeBindings([(V=B)|Bs]) :-
	nl, writeVar(user_output,V), write('='),
	% output not very helpful, should be improved
	writeCurry(B),
	writeBindings(Bs).

% instantiate unbound goal variables by the name of their variable
% in order to print the results in a nicer way:
bindFreeVars([]).
bindFreeVars([(V=B)|Bs]) :-
	(var(B) -> B=V ; true),
	bindFreeVars(Bs).

